#!/usr/bin/python
#-----------------------------------------------------------------------
# File    : surrogates.py
# Contents: Surrogate data generation for point processes/spike trains
# Author  : Christian Borgelt
# History : 2013.11.22 file created
#           2013.12.06 span replaced by variables beg and end
#-----------------------------------------------------------------------
from sys    import stderr
from math   import fmod, floor, ceil
from random import random, triangular, gauss, normalvariate
from random import choice, shuffle

#-----------------------------------------------------------------------
# Constants
#-----------------------------------------------------------------------
oo = float('inf')               # positive infinity as a constant

#-----------------------------------------------------------------------
# Random Distribution Functions (symmetric with mode/mean 0)
#-----------------------------------------------------------------------

def rand_uniform (sigma=1.0):
    '''rand_uniform (sigma=1.0)
Generate a uniformly distributed random number with mean 0.
(This function is identical to rand_rect.)
sigma   spread of the uniform density function
returns a random number in [-sigma,sigma]'''
    return 2*sigma *(random()-0.5)

#-----------------------------------------------------------------------

def rand_rect (sigma=1.0):
    '''rand_rect (sigma=1.0)
Generate a rectangularly distributed random number with mean 0.
(This function is identical to rand_uniform.)
sigma   spread of the rectangular density function
returns a random number in [-sigma,sigma]'''
    return 2*sigma *(random()-0.5)

#-----------------------------------------------------------------------

def rand_triang (sigma=1.0):
    '''rand_triang (sigma=1.0)
Generate a triangularly distributed random number with mode/mean 0.
sigma   spread of the triangle base of the density function
returns a random number in [-sigma,sigma]'''
    return triangular(-sigma,sigma,0)

#-----------------------------------------------------------------------

def rand_gauss (sigma=1.0):
    '''rand_gauss (sigma=1.0)
Generate a normally distributed random number with mode/mean 0.
(This function is slightly faster than rand_normal.)
sigma   standard deviation of the distribution
returns a random number sampled from a Gaussian distribution'''
    return gauss(0,sigma)

#-----------------------------------------------------------------------

def rand_normal (sigma=1.0):
    '''rand_normal (sigma=1.0)
Generate a normally distributed random number with mode/mean 0.
(This function is slightly slower than rand_gauss.)
sigma   standard deviation of the distribution
returns a random number sampled from a normal distribution'''
    return normalvariate(0,sigma)

#-----------------------------------------------------------------------

randfns = [rand_uniform, rand_triang, rand_gauss]
rfndict = { 'u': rand_uniform, 'r': rand_rect,
            't': rand_triang,  'g': rand_gauss, 'n': rand_normal }

#-----------------------------------------------------------------------

def getrandfn (i):
    '''getrandfn (i)
Get random distribution function via an identifier.
i       identifier of the random distribution function
returns a random distribution function'''
    if isinstance(i, (0).__class__):
        return randfns[i] if i >= 0 else len(randfns)
    elif i in rfndict:
        return rfndict[i]
    else: return len(randfns)

#-----------------------------------------------------------------------
# Auxiliary Functions
#-----------------------------------------------------------------------

def getspan (trains):
    '''getspan (trains)
Get (rounded) point/time span of a set of trains.
trains  a list of pairs consisting of an item id and a list of points
returns a pair with minimum and maximum point/time'''
    return (floor(min(min(t) for n,t in trains)),
            ceil (max(max(t) for n,t in trains)))

#-----------------------------------------------------------------------

def adjust (p, beg, end):
    '''adjust (p, beg, end)
Adjust a point/time p w.r.t. a given time span [a,z].
p       point/time to adjust
beg     start of the span to adjust to
end     end   of the span to adjust to
returns the adjusted point/time'''
    p = fmod(p-beg,end-beg)     # wrap round if outside span
    return p +(end if p < 0 else beg)

#-----------------------------------------------------------------------

def surr_ident (trains, rand, beg=-oo, end=oo, delta=0.03):
    '''surr_ident (trains, rand, beg=-oo, end=oo, delta=0.03)
Do not generate surrogate data, merely return original data.
trains  a list of pairs consisting of an item id and a list of points
rand    ignored for this surrogate method
beg     ignored for this surrogate method
end     ignored for this surrogate method
delta   ignored for this surrogate method
returns a list of surrogate spike trains'''
    return trains               # return original spike trains

#-----------------------------------------------------------------------

def surr_random (trains, rand, beg=-oo, end=oo, delta=0.03):
    '''surr_random (trains, rand, beg=-oo, end=oo, delta=0.03)
Generate surrogate data with point/spike time randomization.
trains  a list of pairs consisting of an item id and a list of points
rand    function for random numbers (symmetric with mean 0)
beg     beginning of points (to clamp displaced points/spikes)
end     end       of points (to clamp displaced points/spikes)
delta   ignored for this surrogate method
returns a list of surrogate spike trains'''
    if beg <= -oo or beg >= oo: # get beginning of points
        beg = floor(min(min(t) for n,t in trains))
    if end <= -oo or end >= oo: # get beginning of points
        end = ceil (max(max(t) for n,t in trains))
    s = (end-beg)               # get the span width
    return [(n, [s*random()+beg for x in t])
            for n,t in trains]  # choose new random points/times

#-----------------------------------------------------------------------

def surr_dither (trains, rand, beg=-oo, end=oo, delta=0.03):
    '''surr_dither (trains, rand, beg=-oo, end=oo, delta=0.03)
Generate surrogate data by point/spike time dithering.
trains  a list of pairs consisting of an item id and a list of points
rand    function for random numbers (symmetric with mean 0)
beg     beginning of points (to clamp displaced points/spikes)
end     end       of points (to clamp displaced points/spikes)
delta   ignored for this surrogate method
returns a list of surrogate spike trains'''
    if beg <= -oo or beg >= oo: # get beginning of points
        beg = floor(min(min(t) for n,t in trains))
    if end <= -oo or end >= oo: # get beginning of points
        end = ceil (max(max(t) for n,t in trains))
    return [(n, [adjust(x+rand(),beg,end) for x in t])
            for n,t in trains]  # randomly displace each spike

#-----------------------------------------------------------------------

def surr_shift (trains, rand, beg=-oo, end=oo, delta=0.03):
    '''surr_shift (trains, rand, beg=-oo, end=oo, delta=0.03)
Generate surrogate data by train shifting/dithering.
trains  a list of pairs consisting of an item id and a list of points
rand    function for random numbers (symmetric with mean 0)
beg     beginning of points (to clamp displaced points/spikes)
end     end       of points (to clamp displaced points/spikes)
delta   ignored for this surrogate method
returns a list of surrogate spike trains'''
    if beg <= -oo or beg >= oo: # get beginning of points
        beg = floor(min(min(t) for n,t in trains))
    if end <= -oo or end >= oo: # get beginning of points
        end = ceil (max(max(t) for n,t in trains))
    trains = [(rand(),n,t) for n,t in trains]
    return [(n, [adjust(x+s,beg,end) for x in t])
            for s,n,t in trains]# randomly shift each train

#-----------------------------------------------------------------------

def surr_kernest (trains, rand, beg=-oo, end=oo, delta=0.03):
    '''surr_kernest (trains, rand, beg=-oo, end=oo, delta=0.03)
Generate surrogate data by sampling from a kernel estimate
(random spike drawing; sampling with replacement).
trains  a list of pairs consisting of an item id and a list of points
rand    function for random numbers (symmetric with mean 0)
beg     beginning of points (to clamp displaced points/spikes)
end     end       of points (to clamp displaced points/spikes)
delta   ignored for this surrogate method
returns a list of surrogate spike trains'''
    if beg <= -oo or beg >= oo: # get beginning of points
        beg = floor(min(min(t) for n,t in trains))
    if end <= -oo or end >= oo: # get beginning of points
        end = ceil (max(max(t) for n,t in trains))
    p = [x for n,t in trains for x in t]
    return [(n, [adjust(choice(p)+rand(),beg,end) for x in t])
            for n,t in trains]  # sample from kernel estimate

#-----------------------------------------------------------------------

def surr_permute (trains, rand, beg=-oo, end=oo, delta=0.03):
    '''surr_permute (trains, rand, beg=-oo, end=oo, delta=0.03)
Generate surrogate data by dithering a point/spike permutation.
trains  a list of pairs consisting of an item id and a list of points
rand    function for random numbers (symmetric with mean 0)
beg     beginning of points (to clamp displaced points/spikes)
end     end       of points (to clamp displaced points/spikes)
delta   ignored for this surrogate method
returns a list of surrogate spike trains'''
    if beg <= -oo or beg >= oo: # get beginning of points
        beg = floor(min(min(t) for n,t in trains))
    if end <= -oo or end >= oo: # get beginning of points
        end = ceil (max(max(t) for n,t in trains))
    p = [x for n,t in trains for x in t]
    shuffle(p)                  # shuffle and dither the spike times
    p = [adjust(x+rand(), beg, end) for x in p]
    surr = []; off = 0          # initialize the surrogate
    for n,t in trains:          # traverse the trains
        k = len(t); off += k    # append shuffled sections
        surr.append((n, p[off-k:off]))
    return surr                 # return created surrogate trains

#-----------------------------------------------------------------------

def surr_pblock (trains, rand, beg=-oo, end=oo, delta=0.03):
    '''surr_pblock (trains, rand, beg=-oo, end=oo, delta=0.03)
Generate surrogate data by dithering a point/spike permutation.
trains  a list of pairs consisting of an item id and a list of points
rand    function for random numbers (symmetric with mean 0)
beg     beginning of points (to clamp displaced points/spikes)
end     end       of points (to clamp displaced points/spikes)
delta   block size/time interval in which to permute points/spikes
returns a list of surrogate spike trains'''
    if beg <= -oo or beg >= oo: # get beginning of points
        beg = floor(min(min(t) for n,t in trains))
    if end <= -oo or end >= oo: # get beginning of points
        end = ceil (max(max(t) for n,t in trains))
    p    = sorted([x for n,t in trains for x in t]) +[oo]
    pos  = [0 for t in trains]  # get sorted list of points
    surr = [(n,[]) for n,t in trains]
    l = r = 0; e = beg          # initialize the surrogate trains
    while p[l] < oo:            # while more points to permute
        e += delta              # compute end of next block
        while p[r] < e: r += 1  # collect points in the block
        if r <= l: continue     # check for an empty block
        shuffle(p[l:r])         # shuffle the points in the block
        for k in range(len(trains)):
            t = trains[k][1]    # traverse the trains
            i = j = len(surr[k][1])
            while j < len(t) and t[j] < e:
              j += 1            # find the end of the section
            if j <= i: continue # check for an empty section
            i = l +(j-i)        # compute end of section in points
            surr[k][1].extend([adjust(x+rand(), beg, end)
                               for x in p[l:i]])
            l = i               # append shuffled sections
        if l != r: print("panic!")
    return surr                 # return created surrogate trains

#-----------------------------------------------------------------------

surrfns = [surr_ident, surr_random, surr_dither, surr_shift,
           surr_kernest, surr_permute]
sfndict = {'x': surr_ident,   'i': surr_ident,
           'r': surr_random,  'd': surr_dither,  's': surr_shift,
           'k': surr_kernest, 'p': surr_permute, 'b': surr_pblock}

#-----------------------------------------------------------------------

def getsurrfn (i):
    '''getsurrfn (i)
Get surrogate data generation function via an identifier.
i       identifier of the surrogate data generation function
returns a surrogate data generation function'''
    if isinstance(i, (0).__class__):
        return surrfns[i] if i >= 0 else len(surrfns)
    elif i in sfndict:
        return sfndict[i]
    else: return len(surrfns)
